import os
import random
import tempfile
from os.path import join

import numpy
from arbol import aprint, asection
from dask.array.image import imread
from skimage.color import gray2rgba
from skimage.data import logo, astronaut, camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.video.blend import blend_color_image_sequences


def demo_blend_numpy():
    with NumpyBackend():
        demo_blend(alphas=(0.1, 1, 0.9),
                   scales=(1.0, 0.5, 0.3),
                   translations=((0, 0), (100, 100), (50, 150)),
                   display=True)
        demo_blend(display=True)


def demo_blend_cupy():
    try:
        with CupyBackend():
            demo_blend(alphas=(0.5, 1, 0.9),
                       scales=(1.0, 0.5, 0.3),
                       translations=((0, 0), (100, 100), (50, 150)),
                       display=True)
            demo_blend(display=True)

            return True

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")
        return False


def demo_blend(n=8, alphas=(1, 1, 0.9), scales=None, translations=None, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("Prepare dataset..."):
        image_u = Backend.to_backend(logo())
        image_v = Backend.to_backend(astronaut()[0:500, 0:500])
        image_w = Backend.to_backend(gray2rgba(camera()[0:500, 0:500]))

        # modulate alpha channel:
        image_u[:, 0:256, 3] = 128

        # generate reference 'ground truth' timelapse
        images_u = (image_u.copy() for _ in range(n))
        images_v = (image_v.copy() for _ in range(n))

        # modify each image:
        images_u = (sp.ndimage.shift(image, shift=(random.uniform(-1, 2), random.uniform(-2, 1), 0)) for image in images_u)
        images_v = (sp.ndimage.shift(image, shift=(random.uniform(-2, 1), random.uniform(-2, 3), 0)) for image in images_v)

        # Convert back images to 8 bit:
        images_u = list(image.astype(numpy.uint8) for image in images_u)
        images_v = list(image.astype(numpy.uint8) for image in images_v)

    with tempfile.TemporaryDirectory() as tmpdir:
        aprint('created temporary directory', tmpdir)

        with asection("Save image sequences..."):
            # Create two subfolders for the two image sequences:
            folder_u = join(tmpdir, 'images_u')
            os.makedirs(folder_u)
            folder_v = join(tmpdir, 'images_v')
            os.makedirs(folder_v)

            # Save images from first sequence:
            for i, image_u in enumerate(images_u):
                from PIL import Image
                im = Image.fromarray(Backend.to_numpy(image_u))
                im.save(join(folder_u, f"frame_{i:05}.png"))

            # Save images from second sequence:
            for i, image_v in enumerate(images_v):
                from PIL import Image
                im = Image.fromarray(Backend.to_numpy(image_v))
                im.save(join(folder_v, f"frame_{i:05}.png"))

            # Create png files for image w image_w
            from PIL import Image
            im = Image.fromarray(Backend.to_numpy(image_w))
            path_w = join(tmpdir, f"someimage.png")
            im.save(path_w)

        with asection("Blend sequences..."):
            # Create output folder:
            output_folder = join(tmpdir, 'blend')
            os.makedirs(output_folder)

            # Perform blend:
            blend_color_image_sequences(input_paths=(folder_u, folder_v, path_w),
                                        output_path=output_folder,
                                        modes=('max', 'mean', 'max'),
                                        alphas=alphas,
                                        scales=scales,
                                        translations=translations)

        # load images into dask arrays:
        images_u = imread(os.path.join(folder_u, 'frame_*.png'))
        images_v = imread(os.path.join(folder_v, 'frame_*.png'))
        images_w = imread(path_w)
        images_blend = imread(os.path.join(output_folder, 'frame_*.png'))

        if display:
            from napari import Viewer, gui_qt
            with gui_qt():
                def _c(array):
                    return Backend.to_numpy(array)

                viewer = Viewer(ndisplay=2)
                viewer.add_image(_c(images_u), name='images_u', rgb=True)
                viewer.add_image(_c(images_v), name='images_v', rgb=True)
                viewer.add_image(_c(images_w), name='images_w', rgb=True)
                viewer.add_image(_c(images_blend), name='images_blend', rgb=True)

                viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_blend_cupy():
        demo_blend_numpy()



import json
from typing import Tuple, Union, Sequence

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel


class TranslationRegistrationModel(PairwiseRegistrationModel):

    def __init__(self,
                 shift_vector: Union[Sequence[float], numpy.ndarray],
                 confidence: Union[numpy.ndarray, float] = 1.0,
                 force_numpy: bool = True):

        """ Instantiates a translation registration model

        Parameters
        ----------
        shift_vector : Relative shift between two images
        confidence : registration confidence: a float within [0, 1] which conveys how confident is the registration.
        A value of 0 means no confidence, a value of 1 means perfectly confident.
        force_numpy : when creating this object, you have the option of forcing the use of numpy array instead of the current backend arrays.

        """
        super().__init__()
        xp = Backend.get_xp_module()

        if force_numpy:
            self.shift_vector = Backend.to_numpy(0 if shift_vector is None else shift_vector)
            self.confidence = Backend.to_numpy(0 if confidence is None else confidence)
        else:
            self.shift_vector = xp.asarray(0 if shift_vector is None else shift_vector)
            self.confidence = xp.asarray(0 if confidence is None else confidence)

    def __str__(self):
        return f"TranslationRegistrationModel(shift={self.shift_vector}, confidence={self.confidence})"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.shift_vector == other.shift_vector).all() and (self.confidence == other.confidence).all()
        else:
            return False

    def to_json(self) -> str:
        return json.dumps({'type': 'translation', 'translation': self.shift_vector.tolist(), 'confidence': self.confidence.tolist()})

    def to_numpy(self) -> 'TranslationRegistrationModel':
        self.shift_vector = Backend.to_numpy(self.shift_vector)
        self.confidence = Backend.to_numpy(self.confidence)
        return self

    def integral_shift(self):
        return tuple(int(round(float(shift))) for shift in self.shift_vector)

    def padding(self):
        integral_shift_vector = self.integral_shift()
        padding = tuple(((0, abs(s)) if s > 0 else (abs(s), 0)) for s in integral_shift_vector)
        return padding

    def overall_confidence(self) -> float:
        return float(self.confidence)

    def change_relative_to(self, other) -> float:
        return float(numpy.linalg.norm(Backend.to_numpy(other.shift_vector) - Backend.to_numpy(self.shift_vector)))

    def get_shift_and_confidence(self):
        return self.shift_vector, self.confidence

    def apply(self,
              image,
              integral: bool = True,
              pad: bool = False) -> 'Array':
        """
        Applies the translation model to the given image, possibly by padding the image.

        Parameters
        ----------
        image: image to apply translation to.
        integral: True to snap translation to pixels, False to perform subpixel interpolation.
        pad: True to translate by padding, False for wrapped translation (image shape remains unchanged)

        Returns
        -------
        translated image, with or without padding.

        """
        image = Backend.to_backend(image)

        integral_shift_vector = self.integral_shift()

        if pad:
            xp = Backend.get_xp_module()
            padding = tuple(((0, abs(s)) if s < 0 else (abs(s), 0)) for s in integral_shift_vector)
            image = xp.pad(image, pad_width=padding)
            return image

        else:
            if integral:
                image = numpy.roll(image,
                                   shift=integral_shift_vector,
                                   axis=range(len(integral_shift_vector)))

                return image
            else:
                sp = Backend.get_sp_module()
                image = sp.ndimage.shift(image,
                                         shift=self.shift_vector,
                                         order=1)
                return image

    def apply_pair(self,
                   image_a,
                   image_b,
                   integral: bool = True,
                   pad: bool = False) -> Tuple['Array', 'Array']:
        """
        Applies the translation model to an image pair, possibly by padding the image.

        Parameters
        ----------
        image_a: First image
        image_b: Second image
        integral: True to snap translation to pixels, False to perform subpixel interpolation.
        pad: True to translate by padding, False for wrapped translation (image shape remains unchanged).

        Note: padding implies integral translation.

        Returns
        -------
        Both images registered
        """
        xp = Backend.get_xp_module()

        if pad:
            integral_shift_vector = self.integral_shift()

            padding_a = tuple(((0, abs(s)) if s >= 0 else (abs(s), 0)) for s in integral_shift_vector)
            padding_b = tuple(((0, abs(s)) if s < 0 else (abs(s), 0)) for s in integral_shift_vector)

            image_a = Backend.to_backend(image_a)
            image_b = Backend.to_backend(image_b)

            image_a = xp.pad(image_a, pad_width=padding_a)
            image_b = xp.pad(image_b, pad_width=padding_b)

            return image_a, image_b

        else:
            return image_a, self.apply(image_b, integral=integral, pad=False)



#!/usr/bin/env python

from dexp.optics.psf.standard_psfs import nikon16x08na, olympus20x10na


def demo_standard_psfs():
    """
    Particle scan, focus offset.
    """
    nikon16x08na_psf = nikon16x08na()
    olympus20x10na_psf = olympus20x10na()

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(nikon16x08na_psf, name='nikon16x08na_psf')
        viewer.add_image(olympus20x10na_psf, name='olympus20x10na_psf')


demo_standard_psfs()



from typing import Tuple, Optional

from dexp.processing.backends.backend import Backend


def projection_generator(image,
                         axis_range: Optional[Tuple[int, ...]] = None,
                         projection_type: str = 'mean',
                         nb_axis: int = 2):
    """
    Generates all nD projection of an image. Currently only supports 2D projections.

    Parameters
    ----------
    image: image to compute center of mass of.
    axis_range: Axis range to compute projections, can be: (n) for [0, n[, (m,n) for [m,n[, and (m,n,step) for {i*step | m<=i*step<n & i integer}
    projection_type: Projection type to use when in 'projection' mode: 'mean', 'min', 'max', 'max-min'
    nb_axis: number of axis to project to, currently only supports 2D projections.

    Returns
    -------
    tuples of axis and projections: (u,v,..., projection)

    """
    xp = Backend.get_xp_module()

    if nb_axis != 2:
        raise NotImplementedError(f"Not implemented for nb_axis!=2")

    ndim = image.ndim

    if axis_range is None:
        axis_range = (0, ndim, 1)

    for u in range(*axis_range):
        for v in range(*axis_range):
            if u < v:
                proj_axis = tuple(set(range(*axis_range)).difference({u, v}))
                if projection_type == 'mean':
                    projected_image = xp.mean(image, axis=proj_axis)
                elif projection_type == 'median':
                    projected_image = xp.median(image, axis=proj_axis)
                elif projection_type == 'max':
                    projected_image = xp.max(image, axis=proj_axis)
                elif projection_type == 'min':
                    projected_image = xp.min(image, axis=proj_axis)
                elif projection_type == 'max-min':
                    projected_image = xp.max(image, axis=proj_axis)
                    projected_image -= xp.min(image, axis=proj_axis)
                else:
                    raise ValueError(f"Unknown projection type: {projection_type}")

                projected_image = projected_image.astype(dtype=image.dtype, copy=False)

                yield u, v, projected_image

            elif u < 0 or v < 0 or u >= ndim or v >= ndim:
                raise ValueError(f"Axis range {axis_range} is out-of-bounds for the image of dimension: {ndim}")



import os
from os import listdir
from os.path import exists, isfile, join, isdir

import click
import imageio
import numpy
from arbol.arbol import aprint, asection
from joblib import Parallel, delayed

from dexp.cli.dexp_main import _default_workers_backend
from dexp.processing.backends.backend import Backend


@click.command()
@click.argument('input_paths', nargs=-1)
@click.option('--output_path', '-o', type=str, default=None, help='Output folder for stacked frames.')
@click.option('--orientation', '-or', type=str, default='horiz', help='Stitching mode: horiz, vert', show_default=True)
@click.option('--overwrite', '-w', is_flag=True, help='Force overwrite of output images.', show_default=True)
@click.option('--workers', '-k', type=int, default=-1, help='Number of worker threads to spawn, set to -1 for maximum number of workers', show_default=True)  #
@click.option('--workersbackend', '-wkb', type=str, default=_default_workers_backend, help='What backend to spawn workers with, can be ‘loky’ (multi-process) or ‘threading’ (multi-thread) ', show_default=True)
def stack(input_paths, output_path, orientation, overwrite, workers, workersbackend):
    """ Stacks video vertically or horizontally (video = folder of images).
    """

    if workers <= 0:
        workers = os.cpu_count() // 2

    if output_path is None:
        output_path = input_paths[0] + '_stack'
    elif output_path.startswith('_'):
        output_path = input_paths[0] + output_path

    os.makedirs(output_path, exist_ok=True)

    # collect all image files:
    image_sequences = []
    for input_path in input_paths:

        if isdir(input_path):
            # path is folder:
            pngfiles = [join(input_path, f) for f in listdir(input_path) if isfile(join(input_path, f)) and f.endswith('.png')]
            pngfiles.sort()

        elif isfile(input_path) and (input_path.endswith('png') or input_path.endswith('jpg')):
            # path is image file:
            pngfiles = [input_path, ]

        image_sequences.append(pngfiles)

    # Basic sanity check on the images sequence: we determine the shortest non-one length :
    min_length = min(len(image_sequence) for image_sequence in image_sequences if len(image_sequence) != 1)
    max_length = max(len(image_sequence) for image_sequence in image_sequences if len(image_sequence) != 1)
    if min_length != max_length:
        aprint(f"Not all image sequences have the same non-one length: min:{min_length}, max:{max_length}")

    # Now we broadcast and crop in time:
    _image_sequences = []
    for image_sequence in image_sequences:
        if len(image_sequence) == 1:
            image_sequence = [image_sequence[0], ] * min_length
        elif len(image_sequence) > min_length:
            image_sequence = image_sequence[0:min_length]
        _image_sequences.append(image_sequence)
    image_sequences = _image_sequences
    nb_timepoints = min_length


    def process(tp):

        # Output file:
        filename = f"frame_{tp:05}.png"
        filepath = join(output_path, filename)

        # Write file:
        if overwrite or not exists(filepath):

            stacked_image_array = None

            # collect all images that need to be blended:
            image_paths = list(image_sequence[tp] for image_sequence in image_sequences)

            # Load images:
            images = list(imageio.imread(image_path) for image_path in image_paths)

            if 'horiz' in orientation:
                aprint(f"Stacks frames horizontally.")
                stacked_image_array = numpy.hstack(images)
            elif 'vert' in orientation:
                aprint(f"Stacks frames vertically.")
                stacked_image_array = numpy.vstack(images)

            aprint(f"Writing file: {filename} in folder: {output_path}")
            imageio.imwrite(filepath,
                            Backend.to_numpy(stacked_image_array),
                            compress_level=1)
        else:
            aprint(f"File: {output_path} already exists! use -w option to force overwrite...")


    with asection(f"stacking PNGs: {input_paths}, saving to: {output_path}, orientation: {orientation}"):
        Parallel(n_jobs=workers, backend=workersbackend)(delayed(process)(tp) for tp in range(nb_timepoints))
        aprint("Done!")



import pytest
from skimage.data import binary_blobs
from skimage.filters import gaussian
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.equalise.equalise_intensity import equalise_intensity


def test_equalise_intensity_numpy():
    with NumpyBackend():
        _equalise_intensity()


def test_equalise_intensity_cupy():
    try:
        with CupyBackend():
            _equalise_intensity()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _equalise_intensity(length=128):
    ratio_gt = 2.77

    image_1 = 300 * binary_blobs(length=length, n_dim=3, blob_size_fraction=0.04, volume_fraction=0.01).astype('f4')
    image_1 = gaussian(image_1, sigma=1)
    image_2 = image_1.copy() * ratio_gt

    image_1 = 95 + random_noise(image_1, mode='gaussian', var=0.5, clip=False)
    image_2 = 95 + random_noise(image_2, mode='gaussian', var=0.5, clip=False)

    # from napari import Viewer
    # with napari.gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(image_1, name='image_1')
    #     viewer.add_image(image_2, name='image_2')

    org_image_1, org_image_2 = image_1.copy(), image_2.copy()
    equ_image_1, equ_image_2, corr_ratio = equalise_intensity(image_1, image_2)

    corr_ratio = Backend.to_numpy(corr_ratio)

    print(f" Ratio:{1 / corr_ratio}")

    assert ratio_gt == pytest.approx(1 / corr_ratio, 1e-2)

    # from napari import Viewer
    # with napari.gui_qt():
    #     viewer = Viewer()
    #     viewer.add_image(org_image_1, name='org_image_1', contrast_limits=(0,700))
    #     viewer.add_image(org_image_2, name='org_image_2', contrast_limits=(0,700))
    #     viewer.add_image(equ_image_1, name='equ_image_1', contrast_limits=(0,700))
    #     viewer.add_image(equ_image_2, name='equ_image_2', contrast_limits=(0,700))



from arbol.arbol import aprint

from dexp.datasets.zarr_dataset import ZDataset


def dataset_serve(dataset: ZDataset,
                  host: str = "0.0.0.0",
                  port: int = 8000):
    if not type(dataset) == ZDataset:
        aprint("Cannot serve a non-Zarr dataset!")
        return

    aprint(dataset.info())
    try:
        from simple_zarr_server import serve
        serve(dataset._root_group, host=host, port=port)
    finally:
        # close destination dataset:
        dataset.close()



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def clean_dark_regions(image,
                       threshold: float,
                       size: int = 3,
                       mode: str = 'uniform',
                       in_place: bool = True,
                       internal_dtype=None
                       ):
    """
    Clean Dark Regions

    Filters out noise in dark regions to facilitate compression.


    Parameters
    ----------
    image : image to correct
    threshold : threshold for 'dark' voxel values.
    size : filter size
    mode : cleaning approach: 'none', 'min', 'uniform', and 'median'
    in_place : True if the input image may be modified in-place.
    internal_dtype : internal dtype for computation

    Returns
    -------
    Cleaned image

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype, force_copy=not in_place)

    if mode == 'none':
        filtered = image.copy()
    elif mode == 'min':
        filtered = sp.ndimage.filters.minimum_filter(image, size=size)
    elif mode == 'median':
        filtered = sp.ndimage.filters.median_filter(image, size=size)
    elif mode == 'uniform':
        filtered = sp.ndimage.filters.uniform_filter(image, size=size)
    else:
        raise ValueError(f'Unknown mode: {mode}, only min, median and uniform supported!')

    mask = sp.ndimage.filters.maximum_filter(filtered, size=size) < threshold

    filtered = sp.ndimage.filters.minimum_filter(image, size=size)

    image[mask] = filtered[mask]

    image = image.astype(original_dtype, copy=False)

    return image



import click
from arbol.arbol import aprint, asection

from dexp.processing.utils.mkl_util import set_mkl_threads

_default_store = 'dir'
_default_clevel = 3
_default_codec = 'zstd'
_default_workers_backend = 'threading'

set_mkl_threads()

# def log_uncaught_exceptions(exception_type, exception, tb):
#     import traceback
#     print(''.join(traceback.format_tb(tb)))
#     print('{0}: {1}'.format(exception_type, exception))


# import sys

# sys.excepthook = log_uncaught_exceptions

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group()
def cli():
    aprint("__________________________________________")
    aprint("  DEXP -- Data EXploration & Processing   ")
    aprint("  Dataset processing dexp_commands             ")
    aprint("  Royer lab                               ")
    aprint("__________________________________________")
    aprint("")

    try:
        from dexp.processing.backends.cupy_backend import CupyBackend
        available = CupyBackend.available_devices()
        with asection(f"Available GPU devices: {available}"):
            for device_id in available:
                from cupy.cuda.runtime import CUDARuntimeError
                try:
                    backend = CupyBackend(device_id)
                    aprint(backend)
                except CUDARuntimeError as e:
                    aprint(f"Error while querying available devices: {e}")

    except (ModuleNotFoundError, NotImplementedError):
        aprint("'cupy' module not found! ignored!")

    pass


from dexp.cli.dexp_commands.info import info
from dexp.cli.dexp_commands.check import check
from dexp.cli.dexp_commands.copy import copy
from dexp.cli.dexp_commands.add import add
from dexp.cli.dexp_commands.tiff import tiff
from dexp.cli.dexp_commands.view import view
from dexp.cli.dexp_commands.serve import serve

from dexp.cli.dexp_commands.deskew import deskew
from dexp.cli.dexp_commands.fuse import fuse
from dexp.cli.dexp_commands.deconv import deconv
from dexp.cli.dexp_commands.stabilize import stabilize
from dexp.cli.dexp_commands.isonet import isonet

from dexp.cli.dexp_commands.speedtest import speedtest

from dexp.cli.video_commands.volrender import volrender
from dexp.cli.dexp_commands.projrender import projrender

cli.add_command(info)
cli.add_command(check)
cli.add_command(copy)
cli.add_command(add)
cli.add_command(tiff)
cli.add_command(view)
cli.add_command(serve)

cli.add_command(deskew)
cli.add_command(fuse)
cli.add_command(stabilize)
cli.add_command(deconv)
cli.add_command(isonet)

cli.add_command(speedtest)

cli.add_command(volrender)
cli.add_command(projrender)



import numpy

from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.processing.utils.scatter_gather_i2i import scatter_gather_i2i
from dexp.utils.timeit import timeit


def demo_lr_deconvolution_numpy():
    with NumpyBackend():
        _demo_lr_deconvolution()


def demo_lr_deconvolution_cupy():
    try:
        with CupyBackend():
            _demo_lr_deconvolution()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_lr_deconvolution(length_xy=256):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=False,
                                                                      length_xy=length_xy,
                                                                      zoom=2,
                                                                      length_z_factor=1,
                                                                      background_stength=0,
                                                                      add_offset=False)

    psf = nikon16x08na()
    # psf = olympus20x10na()
    psf = psf.astype(dtype=numpy.float16)
    image = image.astype(dtype=numpy.float16)

    blurry = fft_convolve(image, psf)
    blurry = blurry - blurry.min()
    blurry = blurry / blurry.max()

    max_correction = 2
    iterations = 20

    with timeit("deconvolved"):
        deconvolved = lucy_richardson_deconvolution(blurry, psf, num_iterations=iterations, padding=17, max_correction=max_correction)

    with timeit("lucy_richardson_deconvolution (scatter-gather)"):
        def f(_image):
            return lucy_richardson_deconvolution(_image, psf, num_iterations=iterations, normalise_input=False, padding=17, max_correction=max_correction)

        deconvolved_sg = scatter_gather_i2i(f, blurry, chunks=length_xy // 2, margins=17, normalise=True, to_numpy=True)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(blurry), name='blurry')
        viewer.add_image(_c(psf), name='psf')
        viewer.add_image(_c(deconvolved), name='deconvolved')
        viewer.add_image(_c(deconvolved_sg), name='deconvolved_sg')


if __name__ == "__main__":
    # demo_lr_deconvolution_cupy()
    demo_lr_deconvolution_numpy()



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.model.warp_registration_model import WarpRegistrationModel
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


# TODO: implement numpy version of warp.
# def test_warp_model_numpy():
#     backend = NumpyBackend()
#     _test_warp_model(backend)


def test_warp_model_cupy():
    try:
        with CupyBackend():
            _test_warp_model()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def _test_warp_model(length_xy=128, warp_grid_size=3):
    xp = Backend.get_xp_module()

    _, _, image = generate_nuclei_background_data(add_noise=False,
                                                  length_xy=length_xy,
                                                  length_z_factor=1,
                                                  zoom=2,
                                                  dtype=numpy.float32)
    image = image[0:255, 0:253, 0:251]

    magnitude = 15
    vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3 + (3,))
    confidence = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3)
    # print(f"vector field applied: {vector_field}")

    with timeit("warp"):
        image_warped = warp(image, -vector_field, vector_field_upsampling=8)

    model = WarpRegistrationModel(vector_field=vector_field, confidence=confidence)

    image, image_reg = model.apply_pair(image, image_warped)

    warped_error = xp.mean(xp.absolute(image - image_warped))
    dewarped_error = xp.mean(xp.absolute(image - image_reg))
    print(f"warped_error = {warped_error}")
    print(f"dewarped_error = {dewarped_error}")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', rendering='attenuated mip')
    #     viewer.add_image(_c(image_warped), name='image_warped', colormap='bop purple', blending='additive', visible=False, rendering='attenuated mip')
    #     viewer.add_image(_c(image_reg), name='image_reg', colormap='bop blue', blending='additive', rendering='attenuated mip')
    #     #viewer.camera.ndisplay=3

    assert dewarped_error < warped_error
    assert dewarped_error < 55



from typing import Union, Callable

from matplotlib.cm import get_cmap
from matplotlib.colors import LinearSegmentedColormap, ListedColormap

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def rgb_colormap(image,
                 cmap: Union[str, Callable] = None,
                 bytes: bool = False,
                 internal_dtype=None):
    """
    Takes an image and returns a new image of same shape + an RGB dimension.
    Colors are determined by a provided colormap.


    Parameters
    ----------
    image : Image to project
    cmap: Color map to use, can be a string or a cmap object
    bytes: If true the returned RGB values will be bytes (uint8) between 0 and 255
    internal_dtype: dtype for internal computation

    Returns
    -------
    RGB image of shape (..., 3 or 4).

    """

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = xp.float32

    # move image to current backend:
    image = Backend.to_backend(image, dtype=internal_dtype)

    # set default cmap:
    if cmap is None:
        cmap = 'viridis'

    # normalise color map:
    cmap = _normalise_colormap(cmap)

    if type(cmap) == ListedColormap or type(cmap) == LinearSegmentedColormap:
        listed_cmap: ListedColormap = cmap
        # force initialisation:
        listed_cmap(0)
        rgb_image = _apply_lut(image,
                               listed_cmap._lut,
                               bytes=bytes,
                               N=listed_cmap.N,
                               internal_dtype=internal_dtype)
    else:
        raise NotImplementedError(f"rgb_colormap not yet implemented for cmap type: {type(cmap)}")

    return rgb_image


def _normalise_colormap(cmap):
    if type(cmap) == str:
        cmap = 'cet_rainbow' if cmap == 'rainbow' else cmap
        cmap = 'cet_bmy' if cmap == 'bmy' else cmap
        from colorcet import rgb_to_hex
        rgb_to_hex(0, 0, 0)  # this is a dummy call to prevent elimination of the colorcet import by IDEs
        cmap = get_cmap(cmap)
    elif type(cmap) == LinearSegmentedColormap or type(cmap) == ListedColormap:
        # all good here...
        cmap = cmap
    else:
        raise ValueError(f"Unknown colormap: {cmap}")
    return cmap


def _apply_lut(X,
               lut,
               bytes: bool = False,
               N: int = 256,
               internal_dtype=None):
    """
    Parameters
    ----------
    X : float or int, ndarray or scalar
        The data value(s) to convert to RGBA.
        For floats, X should be in the interval ``[0.0, 1.0]`` to
        return the RGBA values ``X*100`` percent along the Colormap line.
        For integers, X should be in the interval ``[0, Colormap.N)`` to
        return RGBA values *indexed* from the Colormap with index ``X``.
    bytes : bool
        If False (default), the returned RGBA values will be floats in the
        interval ``[0, 1]`` otherwise they will be uint8s in the interval
        ``[0, 255]``.
    N: number of entries in lut.

    Returns
    -------
    Tuple of RGBA values if X is scalar, otherwise an array of
    RGBA values with a shape of ``X.shape + (4, )``.
    """
    xp = Backend.get_xp_module()

    # move image to current backend:
    lut = Backend.to_backend(lut, dtype=internal_dtype)

    xa = xp.array(X, copy=True)
    if not xa.dtype.isnative:
        xa = xa.byteswap().newbyteorder()  # Native byteorder is faster.
    if xa.dtype.kind == "f":
        xa *= N
        # Negative values are out of range, but astype(int) would
        # truncate them towards zero.
        xa[xa < 0] = -1
        # xa == 1 (== N after multiplication) is not out of range.
        xa[xa == N] = N - 1
        # Avoid converting large positive values to negative integers.
        xp.clip(xa, -1, N, out=xa)
        xa = xa.astype(int)

    if bytes:
        lut = (lut * 255).astype(xp.uint8)

    rgba = lut[xa]

    return rgba



from functools import reduce

import numpy
from arbol import aprint

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.filters.sobel_filter import sobel_filter
from dexp.processing.registration.model.translation_registration_model import TranslationRegistrationModel


def register_translation_nd(image_a,
                            image_b,
                            denoise_input_sigma: float = 1.5,
                            gamma: float = 1,
                            log_compression: bool = True,
                            edge_filter: bool = True,
                            max_range_ratio: float = 0.9,
                            decimate: int = 16,
                            quantile: float = 0.999,
                            sigma: float = 1.5,
                            force_numpy: bool = False,
                            internal_dtype=None,
                            _display_phase_correlation: bool = False,
                            ) -> TranslationRegistrationModel:
    """
    Registers two nD images using just a translation-only model.
    This uses a full nD robust phase correlation based approach.

    Note: it is recommended to normalise the images to [0, 1] range before registering them, otherwise there might be precision and overflow issues.

    Parameters
    ----------
    image_a : First image to register
    image_b : Second image to register
    denoise_input_sigma : Uses a Gaussian filter to denoise input images.
    gamma : gamma correction on max projections as a preprocessing before phase correlation.
    log_compression : Applies the function log1p to the images to compress high-intensities (usefull when very (too) bright structures are present in the images, such as beads)
    edge_filter : apply sobel edge filter to input images.
    max_range_ratio : maximal range for correlation.
    decimate : How much to decimate when computing floor level
    quantile : Quantile to use for robust min and max
    sigma : sigma for Gaussian smoothing of phase correlogram
    force_numpy : Forces output model to be allocated with numpy arrays.
    internal_dtype : internal dtype for computation

    _display_phase_correlation : For debugging purposes the phase correlation can be displayed with napari

    Returns
    -------
    Translation-only registration model

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if not image_a.dtype == image_b.dtype:
        raise ValueError("Arrays must have the same dtype")

    if internal_dtype is None:
        internal_dtype = image_a.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = xp.float32

    image_a = Backend.to_backend(image_a, dtype=internal_dtype)
    image_b = Backend.to_backend(image_b, dtype=internal_dtype)

    if denoise_input_sigma is not None and denoise_input_sigma > 0:
        image_a = sp.ndimage.filters.gaussian_filter(image_a, sigma=denoise_input_sigma)
        image_b = sp.ndimage.filters.gaussian_filter(image_b, sigma=denoise_input_sigma)

    if log_compression is not None and log_compression:
        image_a = xp.log1p(image_a)
        image_b = xp.log1p(image_b)

    if gamma is not None and gamma != 1:
        image_a **= gamma
        image_b **= gamma

    if edge_filter is not None and edge_filter:
        image_a = sobel_filter(image_a,
                               exponent=1,
                               normalise_input=False)
        image_b = sobel_filter(image_b,
                               exponent=1,
                               normalise_input=False)

    # Compute the phase correlation:
    raw_correlation = _phase_correlation(image_a, image_b, internal_dtype)
    correlation = raw_correlation

    # Max range is computed from max_range_ratio:
    max_ranges = tuple(int(0.5 * max_range_ratio * s) for s in correlation.shape)
    # print(f"max_ranges={max_ranges}")

    # We estimate the noise floor of the correlation:
    center = tuple(s // 2 for s in correlation.shape)
    empty_region = correlation[tuple(slice(0, c - r) for c, r in zip(center, max_ranges))]
    noise_floor_level = xp.percentile(empty_region.ravel()[::decimate].astype(numpy.float32), q=100 * quantile)
    if xp.isnan(noise_floor_level):
        noise_floor_level = xp.mean(empty_region.ravel()[::decimate])
    # print(f"noise_floor_level={noise_floor_level}")

    # Roll the array and crop it to restrict ourself to the search region:
    correlation = correlation[tuple(slice(max(c - r, 0), min(c + r, s)) for c, r, s in zip(center, max_ranges, correlation.shape))]

    # Use that floor to clip anything below:
    correlation = xp.maximum(correlation, noise_floor_level, out=correlation)
    correlation -= noise_floor_level

    # Denoise cropped correlation image:
    if sigma > 0:
        correlation = sp.ndimage.filters.gaussian_filter(correlation, sigma=sigma, mode='wrap')

    # Use the max as quickly computed proxy for the real center:
    max_correlation_flat_index = xp.argmax(correlation, axis=None)
    rough_shift = xp.unravel_index(max_correlation_flat_index, correlation.shape)
    max_correlation = correlation[rough_shift]

    # Compute the signed shift vector:
    shift_vector = xp.array(tuple(int(rs) - r for rs, r in zip(rough_shift, max_ranges)))

    # Compute confidence:
    masked_correlation = correlation.copy()
    mask_size = tuple(max(8, int(s ** 0.9) // 8) for s in masked_correlation.shape)
    masked_correlation[tuple(slice(rs - s, rs + s) for rs, s in zip(rough_shift, mask_size))] = 0
    background_correlation_max = xp.max(masked_correlation)
    epsilon = 1e-6
    confidence = (max_correlation - background_correlation_max) / (epsilon + max_correlation)

    if _display_phase_correlation:
        # DO NOT DELETE, INSTRUMENTATION CODE FOR DEBUGGING
        from napari import gui_qt, Viewer
        with gui_qt():
            aprint(f"shift = {shift_vector}, confidence = {confidence} ")

            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(image_a), name='image_a')
            viewer.add_image(_c(image_b), name='image_b')
            viewer.add_image(_c(raw_correlation), name='raw_correlation', colormap='viridis')
            viewer.add_image(_c(correlation), name='correlation', colormap='viridis')
            viewer.add_image(_c(masked_correlation), name='masked_correlation', colormap='bop orange', blending='additive')
            viewer.grid.enabled = True
            viewer.grid.shape = (2, 3)

    return TranslationRegistrationModel(shift_vector=shift_vector, confidence=confidence, force_numpy=force_numpy)


def _center_of_mass(image):
    image = Backend.to_backend(image)

    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    try:
        return sp.ndimage.center_of_mass(image)
    except AttributeError:
        # Workaround for the lack of implementation of center_of_mass in cupy
        # TODO: remove this code path once center_of_mass is implemented in cupy!

        normalizer = xp.sum(image)

        if abs(normalizer) > 0:
            grids = numpy.ogrid[[slice(0, i) for i in image.shape]]
            grids = list([Backend.to_backend(grid) for grid in grids])

            results = list([xp.sum(image * grids[dir].astype(float)) / normalizer
                            for dir in range(image.ndim)])

            return tuple(float(f) for f in results)
        else:
            return tuple(s / 2 for s in image.shape)


def _phase_correlation(image_a, image_b,
                       internal_dtype=numpy.float32,
                       epsilon: float = 1e-6,
                       window: float = 0.5):
    xp = Backend.get_xp_module(image_a)

    if window > 0:
        window_axis = tuple(xp.hanning(s) ** window for s in image_a.shape)
        window = reduce(xp.multiply, xp.ix_(*window_axis))
        image_a *= window
        image_b *= window

    G_a = xp.fft.fftn(image_a).astype(numpy.complex64, copy=False)
    G_b = xp.fft.fftn(image_b).astype(numpy.complex64, copy=False)
    conj_b = xp.conj(G_b)
    R = G_a * conj_b
    R /= (xp.absolute(R) + epsilon)
    r = xp.fft.ifftn(R).real.astype(internal_dtype, copy=False)
    r = xp.fft.fftshift(r)
    return r



from arbol import aprint

from dexp.datasets.operations.demo.demo_projrender import _demo_projrender
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend


def test_projrender_numpy():
    with NumpyBackend():
        _demo_projrender(display=False)


def test_stabilize_cupy():
    try:
        with CupyBackend():
            _demo_projrender(display=False)

    except ModuleNotFoundError:
        aprint("Cupy module not found! demo ignored")



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.interpolation.warp import warp
from dexp.utils.timeit import timeit


# def test_warp_2d_numpy():
#     backend = NumpyBackend()
#     _test_warp_2d(backend)


def test_warp_1d_cupy():
    try:
        with CupyBackend():
            _test_warp_1d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _test_warp_1d(grid_size=8):
    xp = Backend.get_xp_module()

    image = numpy.random.uniform(low=0, high=1, size=(128,)).astype(dtype=numpy.float32)
    image = Backend.to_backend(image)

    vector_field = numpy.random.uniform(low=-5, high=+5, size=(grid_size,))

    with timeit("warp"):
        warped = warp(image, vector_field, vector_field_upsampling=4)

    with timeit("dewarp"):
        dewarped = warp(warped, -vector_field, vector_field_upsampling=4)

    error = xp.mean(xp.absolute(image - dewarped))
    print(f"error = {error}")

    # from napari import Viewer, gui_qt
    # with gui_qt():
    #     def _c(array):
    #         return backend.to_numpy(array)
    #
    #     viewer = Viewer()
    #     viewer.add_image(_c(image), name='image')
    #     viewer.add_image(_c(vector_field), name='vector_field')
    #     viewer.add_image(_c(warped), name='warped')
    #     viewer.add_image(_c(dewarped), name='dewarped')

    assert error < 0.3



import gc

import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def lipschitz_continuity_correction(image,
                                    num_iterations: int = 2,
                                    correction_percentile: float = 0.1,
                                    lipschitz: float = 0.1,
                                    max_proportion_corrected: float = 1,
                                    decimation: int = 8,
                                    in_place: bool = True,
                                    internal_dtype=None
                                    ):
    """
    'Lipshitz continuity correction'

    'Broken' pixels on detectors typically blink or are very dim or very bright, in any case they are 'out of context'.
    In many cases they will locally break Lipshitz continuity implied by diffraction limited imaging. Here is a simple
    greedy scheme that starts with the  ost infringing voxels and incrementally filters them using local median filtering.


    Parameters
    ----------

    image : image to correct
    num_iterations : number of iterations
    correction_percentile : percentile of pixels to correct per iteration
    lipschitz : lipschitz continuity constant
    max_proportion_corrected : max proportion of pixels to correct overall
    decimation : decimation for speeding up percentile computation
    in_place : True if the input image may be modified in-place.
    internal_dtype : internal dtype

    Returns
    -------
    Lipschitz corrected image

    """

    xp = Backend.get_xp_module()

    if internal_dtype is None:
        internal_dtype = image.dtype

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = numpy.float32

    original_dtype = image.dtype
    image = Backend.to_backend(image, dtype=internal_dtype, force_copy=not in_place)

    total_number_of_corrections = 0

    for i in range(num_iterations):
        print(f"Iteration {i}")
        # TODO: it is slow to recompute the median filter at each iteration,
        # could be done only once but that's less accurate..
        median, error = _compute_error(image, decimation, lipschitz, internal_dtype)
        gc.collect()
        threshold = xp.percentile(
            error.ravel()[::decimation], q=100 * (1 - correction_percentile)
        )

        mask = error > threshold

        num_corrections = xp.sum(mask)
        print(f"Number of corrections: {num_corrections}")

        if num_corrections == 0:
            break

        proportion = (
                             num_corrections + total_number_of_corrections
                     ) / image.size
        print(
            f"Proportion of corrected pixels: {int(proportion * 100)}% (up to now), versus maximum: {int(max_proportion_corrected * 100)}%) "
        )

        if proportion > max_proportion_corrected:
            break

        image[mask] = median[mask]

        total_number_of_corrections += num_corrections

        gc.collect()

    array = image.astype(original_dtype, copy=False)

    return array


def _compute_error(array, decimation: int, lipschitz: float, dtype=numpy.float16):
    sp = Backend.get_sp_module()
    xp = Backend.get_xp_module()
    array = Backend.to_backend(array, dtype=dtype)
    # we compute the error map:
    median = sp.ndimage.filters.median_filter(array, size=3)
    error = median.copy()
    error -= array
    xp.abs(error, out=error)
    xp.maximum(error, lipschitz, out=error)
    error -= lipschitz
    return median, error



import numpy
from arbol import asection
from skimage.data import binary_blobs
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def generate_nuclei_background_data(length_xy=320,
                                    length_z_factor=4,
                                    zoom=1,
                                    add_noise=True,
                                    background_stength=0.2,
                                    background_scale=0.5,
                                    independent_haze=False,
                                    sphere: bool = False,
                                    radius: float = 0.8,
                                    add_offset=True,
                                    dtype=numpy.float16,
                                    internal_dtype=numpy.float16):
    """

    Parameters
    ----------
    length_xy
    length_z_factor
    zoom
    add_noise
    background_stength
    background_scale
    independent_haze
    sphere
    add_offset
    dtype
    internal_dtype

    Returns
    -------

    """
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    if type(Backend.current()) is NumpyBackend:
        internal_dtype = xp.float32

    if type(Backend.current()) is NumpyBackend:
        dtype = numpy.float32

    with asection("generate blob images"):
        image_gt = binary_blobs(length=length_xy, n_dim=3, blob_size_fraction=0.07, volume_fraction=0.1).astype(internal_dtype)

        if independent_haze:
            background = binary_blobs(length=length_xy, n_dim=3, blob_size_fraction=background_scale, volume_fraction=0.5).astype(internal_dtype)
        else:
            background = image_gt.copy()

    with asection("convert blob images to backend"):
        image_gt = Backend.to_backend(image_gt)
        background = Backend.to_backend(background)

    if sphere:
        with asection("sphere mask"):
            lz, ly, lx = image_gt.shape
            x = xp.linspace(-1, 1, num=lx)
            y = xp.linspace(-1, 1, num=ly)
            z = xp.linspace(-1, 1, num=lz)
            xx, yy, zz = numpy.meshgrid(x, y, z, sparse=True)
            distance = xp.sqrt(xx ** 2 + yy ** 2 + zz ** 2)
            mask = distance < radius
            # f = 0.5*(1 + level**3 / (1 + xp.absolute(level)**3))

            image_gt *= mask

    with asection("prepare high/low image pair"):
        if background_stength > 0:
            background = sp.ndimage.gaussian_filter(background, sigma=15)
            background = background / xp.max(background)

    with asection("generate two views via blending"):
        if background_stength > 0:
            image = background_stength * background
            image += image_gt
        else:
            image = image_gt

    if length_z_factor != 1 or zoom != 1:
        with asection("upscale all dimensions and downscale along z"):
            image_gt = sp.ndimage.zoom(image_gt, zoom=(zoom / length_z_factor, zoom, zoom), order=1)
            background = sp.ndimage.zoom(background, zoom=(zoom / length_z_factor, zoom, zoom), order=1)
            image = sp.ndimage.zoom(image, zoom=(zoom / length_z_factor, zoom, zoom), order=1)

    if add_noise:
        with asection("add noise"):
            image = Backend.to_numpy(image)
            image = random_noise(image, mode='speckle', var=0.5)
            image = Backend.to_backend(image)

    with asection("scale image intensities"):
        zero_level = (1 if add_offset else 0) * xp.random.uniform(95, 95 + (10 if add_noise else 0), size=image_gt.shape)
        zero_level = zero_level.astype(image.dtype, copy=False)
        image *= 300
        image += zero_level
        image = xp.clip(image - 1, 0, None, out=image)

    return image_gt.astype(dtype, copy=False), \
           background.astype(dtype, copy=False), \
           image.astype(dtype, copy=False),



import gc
from functools import reduce
from time import time, sleep

import numpy
from arbol import aprint, asection
from joblib import Parallel, delayed

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.interpolation.warp import warp


def test_cupy_basics():
    try:
        import cupy
        with cupy.cuda.Device(0):
            array = cupy.array([1, 2, 3])
            assert cupy.median(array) == 2
            aprint("\nWorked!")

    except (ModuleNotFoundError, NotImplementedError):
        aprint("Cupy module not found! ignored!")


def test_list_devices():
    try:
        # Test list devices:

        # this will fail if cupy is not available:
        import cupy

        available = CupyBackend.available_devices()
        aprint(f"Available devices: {available}")
        assert len(available) > 0

        for device_id in available:
            with CupyBackend(device_id) as backend:
                aprint(backend)
                xp = Backend.get_xp_module()

                array1 = xp.random.uniform(0, 1, size=(128,) * 3).astype(numpy.float32)
                array2 = xp.random.uniform(0, 1, size=(128,) * 3).astype(numpy.float32)

                array1 = Backend.to_backend(array1, numpy.float32)
                array2 = Backend.to_backend(array2, numpy.float32)

    except ModuleNotFoundError:
        aprint("Cupy module not found! Test passes nevertheless!")


def test_allocation_pool():
    try:

        with CupyBackend(enable_memory_pool=True) as root_backend:
            with CupyBackend(enable_memory_pool=True) as backend:
                gc.collect()
                backend.clear_memory_pool()
                gc.collect()

                xp = backend.get_xp_module()

                import cupy

                in_pool_before = backend.mempool.total_bytes() - backend.mempool.used_bytes()
                aprint(f"in_pool_before={in_pool_before}")

                for i in range(10):
                    array = xp.random.uniform(0, 1, size=(128,) * 3).astype(numpy.float32)

                sleep(2)
                backend.synchronise()

                in_pool_after = backend.mempool.total_bytes() - backend.mempool.used_bytes()
                aprint(f"in_pool_after={in_pool_after}")

                assert in_pool_after > in_pool_before

                backend.clear_memory_pool()

                in_pool_after_clear = backend.mempool.total_bytes() - backend.mempool.used_bytes()
                aprint(f"in_pool_after_clear={in_pool_after_clear}")

                assert in_pool_after_clear <= in_pool_before

                for i in range(10):
                    array = xp.random.uniform(0, 1, size=(128,) * 3).astype(numpy.float32)

                in_pool_after_reallocation = backend.mempool.total_bytes() - backend.mempool.used_bytes()
                aprint(f"in_pool_after_reallocation={in_pool_after_reallocation}")

                assert in_pool_after_reallocation > in_pool_after_clear

            in_pool_after_context = root_backend.mempool.total_bytes() - root_backend.mempool.used_bytes()
            aprint(f"in_pool_after_context={in_pool_after_clear}")

            assert in_pool_after_context == in_pool_after_clear


    except ModuleNotFoundError:
        aprint("Cupy module not found! Test passes nevertheless!")


def test_unified_memory():
    try:

        with CupyBackend(enable_memory_pool=True, enable_unified_memory=True):
            xp = Backend.get_xp_module()

            # a lot of arrays...
            arrays = tuple(xp.random.uniform(0, 1, size=(512,) * 3) for i in range(30))

            sum = reduce((lambda x, y: x + y), arrays)

            aprint(sum)

    except ModuleNotFoundError:
        aprint("Cupy module not found! Test passes nevertheless!")


def test_paralell():
    try:

        def f(id):
            with CupyBackend(id):
                aprint(f"Begin: Job on device #{id}")
                xp = Backend.get_xp_module()
                array = xp.random.uniform(0, 1, size=(128,) * 3).astype(numpy.float32)
                array += 1
                aprint(f"End: Job on device #{id}")

        n_jobs = 2
        Parallel(n_jobs=n_jobs, backend='threading')(delayed(f)(id) for id in range(n_jobs))

    except ModuleNotFoundError:
        aprint("Cupy module not found! Test passes nevertheless!")


def test_paralell_with_exclusive():
    try:
        num_devices = len(CupyBackend.available_devices())
        aprint(f"num_devices = {num_devices}")

        job_duration = 2

        def f(id, device):
            aprint(f"Job #{id} waiting to gain access to device {device} to start ...")
            with CupyBackend(device, exclusive=True):
                with asection(f"Begin: Job #{id} on device #{device}"):
                    xp = Backend.get_xp_module()
                    array = xp.random.uniform(0, 1, size=(128,) * 3).astype(numpy.float32)
                    array += 1
                    sleep(job_duration)
                    aprint(f"End: Job #{id} on device #{device}")

        n_jobs = 2 * num_devices
        aprint(f"n_jobs = {n_jobs}")

        start = time()
        with asection(f"Start jobs"):
            Parallel(n_jobs=n_jobs, backend='threading')(delayed(f)(id, id % num_devices) for id in range(n_jobs))
        elapsed_time = time() - start
        aprint(f"elapsed_time={elapsed_time}")

        # the fact that we have only 'num_devices' available, enforce exclusive access to devices, and spawning more jobs than devices, makes the
        # total elapsed time predictable and testable:
        assert num_devices * job_duration < elapsed_time < (num_devices + 0.5) * job_duration

    except ModuleNotFoundError:
        aprint("Cupy module not found! Test passes nevertheless!")


def test_stress():
    try:
        size = 320

        num_devices = len(CupyBackend.available_devices())
        aprint(f"num_devices = {num_devices}")

        def f(id, device):
            aprint(f"Job #{id} waiting to gain access to device {device} to start ...")
            with CupyBackend(device, exclusive=True, enable_unified_memory=True) as backend:
                with asection(f"Begin: Job #{id} on device #{device}"):
                    xp = Backend.get_xp_module()
                    array = xp.random.uniform(0, 1, size=(size,) * 3)
                    array += xp.random.uniform(0, 1, size=(size,) * 3)

                    # Let's have some texture memory allocated too:
                    vector_field = numpy.random.uniform(low=-5, high=+5, size=(8,) * 3 + (3,))
                    for i in range(10):
                        array = warp(array, vector_field, vector_field_upsampling=4)
                        array += xp.random.uniform(0, 1, size=(size,) * 3)

                    backend.clear_memory_pool()
                    array *= 0.1

                    aprint(f"End: Job #{id} on device #{device}")

        n_jobs = 10 * num_devices
        aprint(f"n_jobs = {n_jobs}")

        with asection(f"Start jobs"):
            Parallel(n_jobs=n_jobs, backend='threading')(delayed(f)(id, id % num_devices) for id in range(n_jobs))


    except ModuleNotFoundError:
        aprint("Cupy module not found! Test passes nevertheless!")



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_warp_3d_numpy():
    try:
        with NumpyBackend():
            _demo_warp_3d()
    except NotImplementedError:
        print("Numpy version not yet implemented")


def demo_warp_3d_cupy():
    try:
        with CupyBackend():
            _demo_warp_3d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_warp_3d(length_xy=256, grid_size=8):
    with timeit("generate data"):
        _, _, image = generate_nuclei_background_data(add_noise=True,
                                                      length_xy=length_xy,
                                                      length_z_factor=1,
                                                      zoom=2,
                                                      dtype=numpy.float32)

    newimage = image[0:512, 0:511, 0:509]
    image = newimage

    print(f"shape={image.shape}")

    magnitude = 15
    vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(grid_size,) * 3 + (3,))

    with timeit("warp"):
        warped = warp(image, vector_field, vector_field_upsampling=4, image_to_backend=True)

    with timeit("dewarped"):
        dewarped = warp(warped, -vector_field, vector_field_upsampling=4, image_to_backend=True)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive', rendering='attenuated_mip')
        viewer.add_image(_c(warped), name='warped', colormap='bop purple', blending='additive', rendering='attenuated_mip')
        viewer.add_image(_c(dewarped), name='dewarped', colormap='bop blue', blending='additive', rendering='attenuated_mip')
        viewer.camera.ndisplay = 3


if __name__ == "__main__":
    demo_warp_3d_cupy()
    # demo_warp_3d_numpy()



import cupy

from dexp.processing.backends._cupy.texture.texture import create_cuda_texture
from dexp.processing.backends.backend import Backend


def _warp_3d_cupy(image,
                  vector_field,
                  mode,
                  block_size: int = 8):
    """

    Parameters
    ----------
    image
    vector_field
    mode
    block_size

    Returns
    -------

    """
    xp = Backend.get_xp_module()

    source = r'''
                extern "C"{
                __global__ void warp_3d(float* warped_image,
                                        cudaTextureObject_t input_image,
                                        cudaTextureObject_t vector_field,
                                        int width, 
                                        int height,
                                        int depth)
                {
                    unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;
                    unsigned int y = blockIdx.y * blockDim.y + threadIdx.y;
                    unsigned int z = blockIdx.z * blockDim.z + threadIdx.z;

                    if (x < width && y < height && z < depth)
                    {
                        // coordinates in coord-normalised vector_field texture:
                        float u = float(x)/width;
                        float v = float(y)/height;
                        float w = float(z)/depth;
                        //printf("(%f,%f,%f)\n", u, v, w);

                        // Obtain linearly interpolated vector at (u,v,w):
                        float4 vector = tex3D<float4>(vector_field, u, v, w);
                        
                        //printf("(%f,%f,%f,%f)\n", vector.x, vector.y, vector.z, vector.w);

                        // Obtain the shifted coordinates of the source voxel, 
                        // flip axis order to match numpy order:
                        float sx = 0.5f + float(x) - vector.z;
                        float sy = 0.5f + float(y) - vector.y;
                        float sz = 0.5f + float(z) - vector.x;

                        // Sample source image for voxel value:
                        float value = tex3D<float>(input_image, sx, sy, sz);

                        //printf("(%f, %f, %f)=%f\n", sx, sy, sz, value);

                        // Store interpolated value:
                        warped_image[z*width*height + y*width + x] = value;

                        //TODO: supersampling would help in regions for which warping misses voxels in the source image,
                        //better: adaptive supersampling would automatically use the vector field divergence to determine where
                        //to super sample and by how much.  
                    }
                }
                }
                '''

    if image.ndim != 3 or vector_field.ndim != 4:
        raise ValueError("image or vector field has wrong number of dimensions!")

    # set up textures:
    input_image_tex, input_image_cudarr = create_cuda_texture(image,
                                                              num_channels=1,
                                                              normalised_coords=False,
                                                              sampling_mode='linear',
                                                              address_mode=mode)

    vector_field = cupy.pad(vector_field, pad_width=((0, 0),) * 3 + ((0, 1),), mode='constant')

    vector_field_tex, vector_field_cudarr = create_cuda_texture(vector_field,
                                                                num_channels=4,
                                                                normalised_coords=True,
                                                                sampling_mode='linear',
                                                                address_mode='clamp')

    # Set up resulting image:
    warped_image = xp.empty(shape=image.shape, dtype=image.dtype)

    # get the kernel, which copies from texture memory
    warp_3d_kernel = cupy.RawKernel(source, 'warp_3d')

    # launch kernel
    depth, height, width = image.shape

    grid_x = (width + block_size - 1) // block_size
    grid_y = (height + block_size - 1) // block_size
    grid_z = (depth + block_size - 1) // block_size
    warp_3d_kernel((grid_x, grid_y, grid_z),
                   (block_size,) * 3,
                   (warped_image, input_image_tex, vector_field_tex, width, height, depth))

    del input_image_tex, input_image_cudarr, vector_field_tex, vector_field_cudarr

    return warped_image



import numpy

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.interpolation.warp import warp
from dexp.processing.registration.model.warp_registration_model import WarpRegistrationModel
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_warp_model_numpy():
    with NumpyBackend():
        warp_model()


def demo_warp_model_cupy():
    try:
        with CupyBackend():
            warp_model()
    except ModuleNotFoundError:
        print("Cupy module not found! demo ignored")


def warp_model(length_xy=320, warp_grid_size=3):
    _, _, image = generate_nuclei_background_data(add_noise=False,
                                                  length_xy=length_xy,
                                                  length_z_factor=1,
                                                  zoom=2,
                                                  dtype=numpy.float32)

    magnitude = 15
    vector_field = numpy.random.uniform(low=-magnitude, high=+magnitude, size=(warp_grid_size,) * 3 + (3,))
    print(f"vector field applied: {vector_field}")

    with timeit("warp"):
        image_warped = warp(image, vector_field, vector_field_upsampling=8)

    model = WarpRegistrationModel(vector_field=-vector_field)

    image, image_reg = model.apply_pair(image, image_warped)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image', colormap='bop orange', blending='additive')
        viewer.add_image(_c(image_warped), name='image_warped', colormap='bop purple', blending='additive', visible=False)
        viewer.add_image(_c(image_reg), name='image_reg', colormap='bop blue', blending='additive')


if __name__ == "__main__":
    demo_warp_model_cupy()
    # demo_warp_model_numpy()



from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.registration.demo.demo_sequence_3d import _register_sequence_3d


# TODO: implement numpy version of warp.
# def test_register_sequence_2d_numpy():
#     with  NumpyBackend():
#         register_sequence_2d()


def test_register_translation_3d_cupy():
    try:
        with CupyBackend():
            register_sequence_3d()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def register_sequence_3d(length_xy=256, n=128):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    image, shifted, stabilised, model = _register_sequence_3d(length_xy=length_xy,
                                                              n=n,
                                                              display=False)

    assert len(model) == image.shape[0]



import click
from arbol.arbol import asection, aprint

from dexp.cli.parsing import _parse_channels
from dexp.datasets.open_dataset import glob_datasets


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--channels', '-c', default=None, help='List of channels, all channels when ommited.')
def check(input_paths, channels):
    """ Checks the integrity of a dataset.
    """

    input_dataset, input_paths = glob_datasets(input_paths)
    channels = _parse_channels(input_dataset, channels)

    with asection(f"checking integrity of datasets {input_paths}, channels: {channels}"):
        result = input_dataset.check_integrity(channels)
        input_dataset.close()

        if not result:
            aprint(f"!!! PROBLEM DETECTED, CORRUPTION LIKELY !!!")
            return

    aprint(f"No problem detected.")
    aprint("Done!")



import click
from arbol.arbol import aprint, asection

from dexp.datasets.open_dataset import glob_datasets
from dexp.datasets.operations.serve import dataset_serve


@click.command()
@click.argument('input_paths', nargs=-1)  # ,  help='input path'
@click.option('--host', '-h', type=str, default="0.0.0.0", help='Host to serve from', show_default=True)
@click.option('--port', '-p', type=int, default=8000, help='Port to serve from', show_default=True)
def serve(input_paths,
          host,
          port):
    """ Serves dataset across network.
    """

    input_dataset, input_paths = glob_datasets(input_paths)

    with asection(f"Serving dataset(s): {input_paths}"):
        dataset_serve(input_dataset,
                      host=host,
                      port=port)

        input_dataset.close()
        aprint("Done!")



import glob
from functools import reduce
from os.path import exists, join
from typing import Sequence

from dexp.datasets.clearcontrol_dataset import CCDataset
from dexp.datasets.joined_dataset import JoinedDataset
from dexp.datasets.zarr_dataset import ZDataset


def glob_datasets(glob_paths: Sequence[str]):
    """
    Opens a joined dataset given a list of path patterns (each following the 'glob' convention).

    Parameters
    ----------
    path: path of dataset

    Returns
    -------

    """

    # Apply glob:
    paths = tuple(glob.glob(glob_path) for glob_path in glob_paths)

    # concatenate list of paths:
    paths = reduce(lambda u, v: u + v, paths)

    # remove empty paths:
    paths = (path.strip() for path in paths)

    # remove empty paths:
    paths = (path for path in paths if len(path) > 0)

    # sort paths:
    paths = sorted(list(paths))

    return open_joined_datasets(paths), paths


def open_joined_datasets(paths: Sequence[str]):
    """
    Opens a joined dataset given a list of paths

    Parameters
    ----------
    path: path of dataset

    Returns
    -------
    JoinedDataset
    """

    if len(paths) == 1:
        # If there is only one dataset, no need to join anything:
        return open_dataset(paths[0])
    else:
        # If there are multiple datasets, we join them into a single dataset:
        datasets = tuple(open_dataset(path) for path in paths)
        dataset = JoinedDataset(datasets)
        return dataset


def open_dataset(path: str):
    """
    Opens a Zarr or ClearControl dataset given a path

    Parameters
    ----------
    path: path of dataset

    Returns
    -------
    dataset

    """

    # remove trailing slash:
    if path.endswith('/'):
        path = path[:-1]

    if path.endswith('.zarr.zip') or path.endswith('.zarr'):
        # we can recognise a Zarr dataset by its extension.
        dataset = ZDataset(path)
    elif exists(join(path, 'stacks')):
        # we can recognise a ClearControl dataset by the presence of a 'stacks' sub folder.
        dataset = CCDataset(path)
    else:
        raise ValueError("Dataset type not recognised, or path incorrect!")

    return dataset



def nd_split_slices(array_shape, chunks, margins=None):
    """
    nd_split_slices

    Parameters
    ----------
    array_shape : tuple
    nb_slices : list
    do_shuffle : tuple
    margins : tuple

    Returns
    -------

    """
    if not array_shape:
        yield ()
        return

    if margins is None:
        margins = (0,) * len(array_shape)

    dim_width = array_shape[-1]

    for outer in nd_split_slices(
            array_shape[:-1], chunks[:-1], margins=margins[:-1]
    ):

        slice_width = chunks[-1]
        slice_margin = margins[-1]

        slice_start_range = list(range(0, dim_width, slice_width))

        for slice_start in slice_start_range:
            start = max(0, slice_start - slice_margin)
            stop = min(slice_start + slice_width + slice_margin, dim_width)
            yield outer + (slice(start, stop, 1),)


def remove_margin_slice(array_shape, slice_with_margin, slice_without_margin):
    """
    Remove_margin_slice

    Parameters
    ----------
    array_shape : tuple
    slice_with_margin : array_like
    slice_without_margin : array_like

    Returns
    -------
    sliced_tuple : tuple

    """
    slice_tuple = tuple(
        slice(max(0, v.start - u.start), min(v.stop - u.start, l), 1)
        for l, u, v in zip(array_shape, slice_with_margin, slice_without_margin)
    )
    return slice_tuple



from arbol import aprint
from skimage.data import camera
from telegram import Update
from telegram.ext import MessageHandler, Filters, CallbackContext, CommandHandler

from dexp.utils.telegram_bot.telegram_bot import TelegramBot


def demo_telegram_bot():
    bot = TelegramBot()

    def start(update: Update, context: CallbackContext) -> None:
        """Send a message when the command /start is issued."""
        update.message.reply_text('Hi!')

    def help_command(update: Update, context: CallbackContext) -> None:
        """Send a message when the command /help is issued."""
        update.message.reply_text('Help!')

    def echo(update: Update, context: CallbackContext) -> None:
        """Echo the user message."""
        aprint(update.message.text)
        update.message.reply_text(update.message.text)

    def image(message: str):
        if 'camera' in message:
            return camera(), 'viridis'

    bot.start()

    bot._add_handler(CommandHandler("start", start))
    bot._add_handler(CommandHandler("help", help_command))
    bot.add_image_handler(image)
    bot._add_handler(MessageHandler(Filters.text & ~Filters.command, echo))

    bot.idle()


demo_telegram_bot()



from random import uniform

from arbol import asection
from skimage.color import gray2rgba
from skimage.data import camera

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.color.time_stamp import insert_time_stamp


def demo_time_stamp_numpy():
    with NumpyBackend():
        demo_time_stamp()


def demo_time_stamp_cupy():
    try:
        with CupyBackend():
            demo_time_stamp()
        return True
    except (ModuleNotFoundError, NotImplementedError):
        print("Cupy module not found! ignored!")
        return False


def demo_time_stamp(n=32, display=True):
    xp = Backend.get_xp_module()
    sp = Backend.get_sp_module()

    with asection("generate data"):
        image = Backend.to_backend(gray2rgba(camera()))

        # generate reference 'ground truth' timelapse
        images = (image.copy() for _ in range(n))

        # modify each image:
        images = (sp.ndimage.shift(image, shift=(uniform(-1, 2), uniform(-2, 1), 0)) for image in images)

        # Convert back images to 8 bit:
        images = list(image.astype(xp.uint8) for image in images)

    with asection("Apply time stamp..."):
        images_with_time_stamps = []
        for tp, image in enumerate(images):
            image_with_time_stamp = insert_time_stamp(image=image,
                                                      time_point_index=tp,
                                                      nb_time_points=len(images),
                                                      start_time=0,
                                                      time_interval=20 / 60.0,
                                                      unit='min',
                                                      translation='top_right',
                                                      color=(1, 1, 1, 1),
                                                      alpha=1.0
                                                      )
            images_with_time_stamps.append(image_with_time_stamp)

        # Convert sequences to arrays:
        # images = xp.stack(images)
        images_with_time_stamp_tr = xp.stack(images_with_time_stamps)

    if display:
        from napari import Viewer, gui_qt
        with gui_qt():
            def _c(array):
                return Backend.to_numpy(array)

            viewer = Viewer()
            viewer.add_image(_c(images_with_time_stamp_tr), name='images_with_time_stamp_tr', rgb=True)
            viewer.grid.enabled = True


if __name__ == "__main__":
    if not demo_time_stamp_cupy():
        demo_time_stamp_numpy()



import numpy
import pytest
from skimage.data import binary_blobs

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.numpy_backend import NumpyBackend


def test_numpy_backend():
    array = binary_blobs(length=100, n_dim=3, blob_size_fraction=0.01, volume_fraction=0.01).astype(numpy.float32)

    with NumpyBackend():
        array_b = Backend.to_backend(array, numpy.float64)
        array_r = Backend.to_numpy(array_b, numpy.float32)

        assert pytest.approx(array, rel=1e-5) == array_r



import pytest
from skimage.data import binary_blobs
from skimage.filters import gaussian
from skimage.util import random_noise

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.equalise.equalise_intensity import equalise_intensity
from dexp.utils.timeit import timeit


def demo_equalise_intensity_numpy():
    with NumpyBackend():
        _equalise_intensity()


def demo_equalise_intensity_cupy():
    try:
        with CupyBackend():
            _equalise_intensity()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _equalise_intensity(length=256):
    with timeit("generate demo dataset"):
        ratio_gt = 2.77

        image_1 = 300 * binary_blobs(length=length, n_dim=3, blob_size_fraction=0.04, volume_fraction=0.01).astype('f4')
        image_1 = gaussian(image_1, sigma=1)
        image_2 = image_1.copy() * ratio_gt

        image_1 = 95 + 30 * random_noise(image_1, mode='gaussian', var=0.2, clip=False)
        image_2 = 95 + 30 * random_noise(image_2, mode='gaussian', var=0.2, clip=False)

    org_image_1, org_image_2 = image_1.copy(), image_2.copy()
    with timeit("equalise intensity"):
        equ_image_1, equ_image_2, corr_ratio = equalise_intensity(image_1, image_2)

    corr_ratio = Backend.to_numpy(corr_ratio)

    print(f" Ratio:{1 / corr_ratio}")

    assert ratio_gt == pytest.approx(1 / corr_ratio, 1e-2)

    from napari import Viewer
    import napari
    with napari.gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image_1), name='image_1')
        viewer.add_image(_c(image_2), name='image_2')
        viewer.add_image(_c(org_image_1), name='org_image_1', contrast_limits=(0, 700))
        viewer.add_image(_c(org_image_2), name='org_image_2', contrast_limits=(0, 700))
        viewer.add_image(_c(equ_image_1), name='equ_image_1', contrast_limits=(0, 700))
        viewer.add_image(_c(equ_image_2), name='equ_image_2', contrast_limits=(0, 700))


if __name__ == "__main__":
    demo_equalise_intensity_numpy()
    demo_equalise_intensity_cupy()
